import copy
from collections import OrderedDict

from ..protocol import Protocol, BASE_TYPE_SIZES, COLLECTION_TYPES
from ..writer import Writer
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "Go"


class GoWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".go"

    def __init__(self, p: Protocol):
        # because of how Go handles making methods/members public, we need to
        #   do some rewriting on the protocol here
        def publicize(s: str):
            return s[:1].upper() + s[1:]

        p2 = copy.deepcopy(p)
        nstructs = OrderedDict()
        for s in p2.structs:
            k = publicize(s)
            v = []
            for vdata in p2.structs[s]:
                v.append( (publicize(vdata[0]), vdata[1]) )
            nstructs[k] = v
        p2.structs = nstructs

        nmessages = OrderedDict()
        for m in p2.messages:
            k = publicize(m)
            v = []
            for vdata in p2.messages[m]:
                v.append( (publicize(vdata[0]), vdata[1]) )
            nmessages[k] = v
        p2.messages = nmessages

        super().__init__(protocol=p2, tab="\t")

        self.type_mapping["byte"] = "byte"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "uint16"
        self.type_mapping["int16"] = "int16"
        self.type_mapping["uint32"] = "uint32"
        self.type_mapping["int32"] = "int32"
        self.type_mapping["uint64"] = "uint64"
        self.type_mapping["int64"] = "int64"
        self.type_mapping["float"] = "float32"
        self.type_mapping["double"] = "float64"


    def deserializer(self, var_type: str, var_name: str, parent: str = None) -> list[str]:
        if parent == None:
            pref = ""
            this = var_name
        else:
            pref = f"{parent}."
            this = f"{pref}{var_name}"
        label = var_name
        if label.endswith("[i]"):
            label = "i"

        err_panic = [
            "if err != nil {",
            self.tab + "panic(err)",
            "}",
        ]

        if self.protocol.is_simple(var_type):
            return [
                f"err = binary.Read(data, binary.LittleEndian, &{pref}{var_name})",
                *err_panic
            ]
        elif var_type in self.protocol.structs or var_type in self.protocol.messages:
            fields: list[tuple[str,str]] = None
            if var_type in self.protocol.structs:
                fields = self.protocol.structs[var_type]
            elif var_type in self.protocol.messages:
                fields = self.protocol.messages[var_type]
            output: list[str] = []
            for vn, vt in fields:
                output += self.deserializer(vt, vn, this)
            return output
        elif var_type == "string":
            return [
                f"err = readString(data, &{pref}{var_name})",
                "if err != nil {",
                f"{self.tab}panic(err)",
                "}",
            ]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"var {label}Len uint32",
                f"err = binary.Read(data, binary.LittleEndian, &{label}Len)",
                *err_panic,
                f"{pref}{var_name} = make([]{interior}, {label}Len)",
                f"for {var_name}_i := (uint32)(0); {var_name}_i < {label}Len; {var_name}_i++ {{"
            ]
            out += [
                self.tab + deser for deser in self.deserializer(interior, f"{var_name}[{var_name}_i]", parent)
            ]
            out += ["}"]
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not deserializable yet.")


    def serializer(self, var_type: str, var_name: str, parent: str = None) -> list[str]:
        if parent == None:
            pref = ""
            ptr = ""
            this = "output"
        else:
            pref = f"{parent}."
            ptr = "&"
            this = f"{pref}{var_name}"
        label = var_name
        if label.endswith("[i]"):
            label = "i"

        if self.protocol.is_simple(var_type):
            return [f"binary.Write(data, binary.LittleEndian, {ptr}{pref}{var_name})"]
        elif var_type in self.protocol.structs or var_type in self.protocol.messages:
            fields: list[tuple[str,str]] = None
            if var_type in self.protocol.structs:
                fields = self.protocol.structs[var_type]
            elif var_type in self.protocol.messages:
                fields = self.protocol.messages[var_type]
            output: list[str] = []
            for vn, vt in fields:
                output += self.serializer(vt, vn, this)
            return output
        elif var_type == "string":
            return [f"writeString(data, {ptr}{pref}{var_name})"]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"{label}Len := (uint32)(len({pref}{var_name}))",
                f"binary.Write(data, binary.LittleEndian, {label}Len)",
                f"for {var_name}_i := (uint32)(0); {var_name}_i < {label}Len; {var_name}_i++ {{",
            ]
            out += [
                self.tab + deser for deser in self.serializer(interior, f"{var_name}[{var_name}_i]", parent)
            ]
            out += ["}"]
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not serializable yet.")


    def gen_struct(self, s: tuple[str, list[tuple[str,str]]]):
        is_message = False
        if s[0] in self.protocol.messages:
            is_message = True

        self.write_line()
        self.write_line(f"type {s[0]} struct {{")
        self.indent_level += 1

        for var_name, var_type in s[1]:
            if var_type[0] == "[" and var_type[-1] == "]":
                self.write_line(f"{var_name} []{self.get_var(var_type[1:-1])}")
            else:
                self.write_line(f"{var_name} {self.get_var(var_type)}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if is_message:
            self.write_line(f"func {s[0]}FromBytes (data io.Reader) (msg *{s[0]}) {{")
            self.indent_level += 1
            self.write_line("defer func() {")
            self.indent_level += 1
            self.write_line("if r := recover(); r != nil {")
            self.indent_level += 1
            self.write_line("msg = nil")
            self.indent_level -= 1
            self.write_line("}")
            self.indent_level -= 1
            self.write_line("}()")
            self.write_line("var err error")
            self.write_line(f"ret := {s[0]}{{}}")
            [self.write_line(s) for s in self.deserializer(s[0], "ret")]
            self.write_line()
            self.write_line("return &ret")
        else:
            self.write_line(f"func {s[0]}FromBytes (data io.Reader, input *{s[0]}) {{")
            self.indent_level += 1
            self.write_line("var err error")
            [self.write_line(s) for s in self.deserializer(s[0], "input")]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        if is_message:
            self.write_line(f"func (output {s[0]}) WriteBytes (data io.Writer, tag bool) {{")
            self.indent_level += 1
            self.write_line("if tag {")
            self.indent_level += 1
            self.write_line(f"binary.Write(data, binary.LittleEndian, {s[0]}Type)")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line(f"func (output {s[0]}) WriteBytes (data io.Writer) {{")
            self.indent_level += 1
        [self.write_line(s) for s in self.serializer(s[0], "output")]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def gen_measurement(self, s: tuple[str, list[tuple[str,str]]], accessor_prefix: str = "") -> tuple[list[str], int]:
        lines: list[str] = []

        accum = 0
        if self.protocol.is_simple(s[0]):
            lines.append(f"return {self.protocol.calculate_size(s[0])}")
        else:
            size_init = "size := 0"
            lines.append(size_init)

            for var_name, var_type in s[1]:
                if self.protocol.is_simple(var_type):
                    accum += self.protocol.calculate_size(var_type)
                else:
                    if var_type == "string":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"size += len({accessor_prefix}{var_name})")
                    elif var_type == "[string]":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"for _, s := range {accessor_prefix}{var_name} {{")
                        lines.append(f"{self.tab}size += {BASE_TYPE_SIZES['uint32']} + len(s)")
                        lines.append("}")
                    elif var_type[0] == "[" and var_type[-1] == "]":
                        listed_var_type = var_type[1:-1]
                        if self.protocol.is_simple(listed_var_type):
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"size += len({accessor_prefix}{var_name}) * {self.protocol.calculate_size(listed_var_type)}")
                        else:
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"for _, c := range {accessor_prefix}{var_name} {{")
                            clines, caccum = self.gen_measurement((var_type, self.protocol.structs[listed_var_type]), f"c.")
                            if clines[0] == size_init:
                                clines = clines[1:]
                            clines.append(f"size += {caccum}")
                            lines += [f"{self.tab}{l}" for l in clines]
                            lines.append("}")
                    else:
                        clines, caccum = self.gen_measurement((var_type, self.protocol.structs[var_type]), f"{accessor_prefix}{var_name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_message(self, m: tuple[str, list[tuple[str,str]]]):
        self.gen_struct(m)

        self.write_line(f"func (output {m[0]}) GetMessageType() MessageType {{")
        self.indent_level += 1
        self.write_line(f"return {m[0]}Type")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line(f"func (output {m[0]}) GetSizeInBytes() int {{")
        self.indent_level +=1
        measure_lines, accumulator = self.gen_measurement(m, "output.")
        [self.write_line(s) for s in measure_lines]
        if accumulator > 0:
            self.write_line(f"size += {accumulator}")
        if len(measure_lines) > 1:
            self.write_line(f"return size")
        self.indent_level -=1
        self.write_line("}")



    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}.")
        self.write_line(f"// Do not edit directly.")
        self.write_line()
        if self.protocol.namespace:
            self.write_line(f"package {self.protocol.namespace}")
        else:
            self.write_line("package main")
        self.write_line()
        self.write_line("import (")
        self.indent_level += 1
        self.write_line("\"encoding/binary\"")
        self.write_line("\"io\"")
        self.indent_level -= 1
        self.write_line(")")
        self.write_line()

        msg_types = [mt for mt in self.protocol.messages.keys()]

        self.write_line("type MessageType byte")
        self.write_line("const (")
        self.indent_level += 1
        [self.write_line(f"{k}Type MessageType = {i+1}") for i, k in enumerate(msg_types)]
        self.indent_level -= 1
        self.write_line(")")
        self.write_line()

        self.write_line("type Message interface {")
        self.indent_level += 1
        self.write_line("GetMessageType() MessageType")
        self.write_line("WriteBytes(data io.Writer, tag bool)")
        self.write_line("GetSizeInBytes() int")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("func ProcessRawBytes (data io.Reader) []Message {")
        self.indent_level += 1
        self.write_line("var msgList []Message")
        self.write_line("var err error")
        self.write_line("for err != io.EOF {")
        self.indent_level += 1
        self.write_line("var msgType MessageType")
        self.write_line("err = binary.Read(data, binary.LittleEndian, &msgType)")
        self.write_line("if err == io.EOF {")
        self.indent_level += 1
        self.write_line("break")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("switch msgType {")
        for msg_type in msg_types:
            self.write_line(f"case {msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList = append(msgList, {msg_type}FromBytes(data))")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgList = append(msgList, nil)")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if msgList[len(msgList)-1] == nil {")
        self.indent_level += 1
        self.write_line("break")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("func readString(data io.Reader, str *string) error {")
        self.indent_level += 1
        self.write_line("var len uint32")
        self.write_line("binary.Read(data, binary.LittleEndian, &len)")
        self.write_line("sbytes := make([]byte, len)")
        self.write_line("err := binary.Read(data, binary.LittleEndian, &sbytes)")
        self.write_line("if err != nil {")
        self.indent_level += 1
        self.write_line("panic(err)")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("*str = string(sbytes)")
        self.write_line("return err")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("func writeString(data io.Writer, str *string) {")
        self.indent_level += 1
        self.write_line("strLen := (uint32)(len(*str))")
        self.write_line("binary.Write(data, binary.LittleEndian, strLen)")
        self.write_line("io.WriteString(data, *str)")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for s in self.protocol.structs.items():
            self.gen_struct(s)

        for m in self.protocol.messages.items():
            self.gen_message(m)

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



import string
from collections import OrderedDict

import toml

BASE_TYPE_SIZES: dict[str, int] = {
    "byte":   1,
    "bool":   1,
    "uint16": 2,
    "int16":  2,
    "uint32": 4,
    "int32":  4,
    "uint64": 8,
    "int64":  8,
    "float":  4,
    "double": 8,
}

COLLECTION_TYPES: list[str] = [
    "list",
    "string",
]

def _contains_whitespace(s: str) -> bool:
    return True in [c in s for c in string.whitespace]

class Protocol():
    def __init__(self, filename: str):
        self.namespace: str = None
        self.structs: OrderedDict = OrderedDict()
        self.messages: OrderedDict = OrderedDict()

        protocol_data = toml.load(filename)

        if "meta" in protocol_data and "namespace" in protocol_data["meta"]:
            if _contains_whitespace(protocol_data["meta"]["namespace"]):
                raise ValueError(f"Namespace cannot contain whitespace: '{protocol_data['meta']['namespace']}'")
            self.namespace = protocol_data["meta"]["namespace"]

        if "structs" in protocol_data:
            for struct_data in protocol_data["structs"]:
                if "_name" not in struct_data:
                    raise ValueError(f"Missing _name on {struct_data}")
                if struct_data["_name"] in self.structs:
                    raise ValueError(f"Duplicate _name on {struct_data}")
                if _contains_whitespace(struct_data["_name"]):
                    raise ValueError(f"Struct _name cannot contain whitespace: '{struct_data['_name']}'")
                self.structs[struct_data["_name"]] = []
            for struct in protocol_data["structs"]:
                struct_name = struct["_name"]
                for var_name, var_type in struct.items():
                    if (var_name[0] == "_"):
                        continue
                    if _contains_whitespace(var_name):
                        raise ValueError(f"Member name cannot contain whitespace: '{var_name}'")
                    if not self.verify_type_name(var_type):
                        raise NotImplementedError(f"No type called {var_type} (definition: {struct_name})")
                    self.structs[struct_name].append((var_name, var_type))

        if "messages" in protocol_data:
            for message_data in protocol_data["messages"]:
                if "_name" not in message_data:
                    raise ValueError(f"Missing _name on {message_data}")
                if message_data["_name"] in self.messages:
                    raise ValueError(f"Duplicate _name on {message_data}")
                if _contains_whitespace(message_data["_name"]):
                    raise ValueError(f"Struct _name cannot contain whitespace: '{message_data['_name']}'")
                self.messages[message_data["_name"]] = []
            for message in protocol_data["messages"]:
                message_name = message["_name"]
                for var_name, var_type in message.items():
                    if (var_name[0] == "_"):
                        continue
                    if _contains_whitespace(var_name):
                        raise ValueError(f"Member name cannot contain whitespace: '{var_name}'")
                    if not self.verify_type_name(var_type):
                        raise NotImplementedError(f"No type called {var_type} (message: {message_name})")
                    self.messages[message_name].append((var_name, var_type))

            if len(protocol_data["messages"]) > 255:
                raise ValueError("Cannot, at present, have more than 255 types of messages. Sorry. :(")

    def __str__(self) -> str:
        return f"Protocol (namespace: {self.namespace}, {len(self.structs)} structs, {len(self.messages)} messages)"


    def verify_type_name(self, type_name) -> bool:
        if type_name[0] == "[" and type_name[-1] == "]":
            type_name = type_name[1:-1]

        if type_name in BASE_TYPE_SIZES:
            return True
        if type_name in COLLECTION_TYPES:
            return True
        if type_name in self.structs:
            return True
        return False

    def is_simple(self, var_type: str) -> bool:
        if var_type in BASE_TYPE_SIZES:
            return True
        elif var_type in COLLECTION_TYPES:
            return False
        elif var_type[0] == "[" and var_type[-1] == "]":
            return False
        elif var_type in self.structs or var_type in self.messages:
            datums: list[tuple[str,str]] = None
            if var_type in self.structs:
                datums = self.structs[var_type]
            else:
                datums = self.messages[var_type]
            for _, vt in datums:
                if not self.is_simple(vt):
                    return False
            return True
        else:
            raise NotImplementedError(f"Can't determine simplicity of {var_type}.")

    def calculate_size(self, var_type: str) -> int:
        if self.is_simple(var_type):
            if var_type in BASE_TYPE_SIZES:
                return BASE_TYPE_SIZES[var_type]
            else:
                size = 0
                if var_type in self.structs:
                    datums = self.structs[var_type]
                else:
                    datums = self.messages[var_type]
                for _, vt in datums:
                    size += self.calculate_size(vt)
                return size
        else:
            return -1



from ..protocol import Protocol, BASE_TYPE_SIZES
from ..writer import Writer
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "CSharp"


class CSharpWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".cs"

    def __init__(self, p: Protocol):
        super().__init__(protocol=p, tab="    ")

        self.type_mapping["byte"] = "byte"
        self.type_mapping["bool"] = "bool"
        self.type_mapping["uint16"] = "ushort"
        self.type_mapping["int16"] = "short"
        self.type_mapping["uint32"] = "uint"
        self.type_mapping["int32"] = "int"
        self.type_mapping["uint64"] = "ulong"
        self.type_mapping["int64"] = "long"
        self.type_mapping["float"] = "float"
        self.type_mapping["double"] = "double"


    def deserializer(self, var_type: str, var_name: str, parent: str = "this") -> list[str]:
        if parent:
            pref = parent + "."
        else:
            pref = ""
        label = var_name
        if label.endswith("_i]"):
            label = "i_"
        if var_type in BASE_TYPE_SIZES:
            func = None
            if var_type == "byte":
                func = "ReadByte"
            elif var_type == "bool":
                func = "ReadBoolean"
            elif var_type == "uint16":
                func = "ReadUInt16"
            elif var_type == "int16":
                func = "ReadInt16"
            elif var_type == "uint32":
                func = "ReadUInt32"
            elif var_type == "int32":
                func = "ReadInt32"
            elif var_type == "uint64":
                func = "ReadUInt64"
            elif var_type == "int64":
                func = "ReadInt64"
            elif var_type == "float":
                func = "ReadSingle"
            elif var_type == "double":
                func = "ReadDouble"
            else:
                raise NotImplementedError(f"Type {var_type} not deserializable yet.")
            return [f"{pref}{var_name} = br.{func}();"]
        elif var_type == "string":
            return [
                f"uint {label}Length = br.ReadUInt32();",
                f"byte[] {label}Buffer = br.ReadBytes((int){label}Length);",
                f"{pref}{var_name} = System.Text.Encoding.UTF8.GetString({label}Buffer);",
            ]
        elif var_type in self.protocol.structs:
            return [
                f"{pref}{var_name} = {var_type}.FromBytes(br);"
            ]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"uint {var_name}Length = br.ReadUInt32();",
                f"{pref}{var_name} = new {self.get_var(interior)}[{var_name}Length];",
                f"for (int {var_name}_i = 0; {var_name}_i < {var_name}Length; {var_name}_i++)",
                "{",
            ]
            out += [
                self.tab + deser for deser in self.deserializer(
                    interior, f"{var_name}[{var_name}_i]", parent
                )
            ]
            out += ["}"]
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not deserializable yet.")


    def serializer(self, var_type: str, var_name: str, parent: str = "this") -> list[str]:
        if parent:
            pref = parent + "."
        else:
            pref = ""
        if var_type in BASE_TYPE_SIZES:
            return [f"bw.Write({pref}{var_name});"]
        elif var_type == "string":
            return [
                f"byte[] {var_name}Buffer = System.Text.Encoding.UTF8.GetBytes({pref}{var_name});",
                f"bw.Write((uint){var_name}Buffer.Length);",
                f"bw.Write({var_name}Buffer);",
            ]
        elif var_type in self.protocol.structs:
            return [
                f"{pref}{var_name}.WriteBytes(bw);"
            ]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"bw.Write((uint){pref}{var_name}.Length);",
                f"foreach ({self.get_var(interior)} el in {pref}{var_name})",
                "{"
            ]
            out += [self.tab + ser for ser in self.serializer(interior, "el", None)]
            out += "}"
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not serializable yet.")


    def gen_measurement(self, s: tuple[str, list[tuple[str,str]]], accessor_prefix: str = "") -> tuple[list[str], int]:
        lines: list[str] = []

        accum = 0
        if self.protocol.is_simple(s[0]):
            lines.append(f"return {self.protocol.calculate_size(s[0])};")
        else:
            size_init = "int size = 0;"
            lines.append(size_init)

            for var_name, var_type in s[1]:
                if self.protocol.is_simple(var_type):
                    accum += self.protocol.calculate_size(var_type)
                else:
                    if var_type == "string":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"size += System.Text.Encoding.UTF8.GetBytes({accessor_prefix}{var_name}).Length;")
                    elif var_type == "[string]":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"foreach (var s in {accessor_prefix}{var_name})")
                        lines.append("{")
                        lines.append(f"{self.tab}size += {BASE_TYPE_SIZES['uint32']} + System.Text.Encoding.UTF8.GetBytes(s).Length;")
                        lines.append("}")
                    elif var_type[0] == "[" and var_type[-1] == "]":
                        listed_var_type = var_type[1:-1]
                        if self.protocol.is_simple(listed_var_type):
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"size += {accessor_prefix}{var_name}.Length * {self.protocol.calculate_size(listed_var_type)};")
                        else:
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"foreach (var s in {accessor_prefix}{var_name})")
                            lines.append("{")
                            clines, caccum = self.gen_measurement((var_type, self.protocol.structs[listed_var_type]), f"s.")
                            if clines[0] == size_init:
                                clines = clines[1:]
                            clines.append(f"size += {caccum};")
                            lines += [f"{self.tab}{l}" for l in clines]
                            lines.append("}")
                    else:
                        clines, caccum = self.gen_measurement((var_type, self.protocol.structs[var_type]), f"{accessor_prefix}{var_name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, s: tuple[str, list[tuple[str,str]]]):
        is_message = False
        if (s[0] in self.protocol.messages):
            is_message = True
        if is_message:
            self.write_line(f"public class {s[0]} : Message")
        else:
            self.write_line(f"public class {s[0]}")
        self.write_line("{")
        self.indent_level += 1

        for var_name, var_type in s[1]:
            if var_type[0] == "[" and var_type[-1] == "]":
                self.write_line(f"public {self.get_var(var_type[1:-1])}[] {var_name};")
            else:
                self.write_line(f"public {self.get_var(var_type)} {var_name};")

        if is_message:
            self.write_line()
            self.write_line(f"public override MessageType GetMessageType() {{ return MessageType.{s[0]}Type; }}")
            self.write_line()
            self.write_line("public override int GetSizeInBytes()")
            self.write_line("{")
            self.indent_level +=1
            measure_lines, accumulator = self.gen_measurement(s, "this.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator};")
            if len(measure_lines) > 1:
                self.write_line(f"return size;")
            self.indent_level -=1
            self.write_line("}")

        self.write_line()
        override = ""
        tag = ""
        if is_message:
            override = "override "
            tag = ", bool tag"
        self.write_line(f"public {override}void WriteBytes(BinaryWriter bw{tag})")
        self.write_line("{")
        self.indent_level += 1
        if is_message:
            self.write_line("if (tag)")
            self.write_line("{")
            self.indent_level += 1
            self.write_line(f"bw.Write((byte)MessageType.{s[0]}Type);")
            self.indent_level -= 1
            self.write_line("}")
        for var_name, var_type in s[1]:
            [self.write_line(s) for s in self.serializer(var_type, var_name)]
        self.indent_level -= 1
        self.write_line("}")

        self.write_line()
        self.write_line(f"public static {s[0]} FromBytes(BinaryReader br)")
        self.write_line("{")
        self.indent_level += 1
        if is_message:
            self.write_line("try")
            self.write_line("{")
            self.indent_level += 1
        self.write_line(f"{s[0]} n{s[0]} = new {self.get_var(s[0])}();")
        for var_name, var_type in s[1]:
            [self.write_line(s) for s in self.deserializer(var_type, var_name, f"n{s[0]}")]
        self.write_line(f"return n{s[0]};")
        if is_message:
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("catch (System.IO.EndOfStreamException)")
            self.write_line("{")
            self.indent_level += 1
            self.write_line("return null;")
            self.indent_level -= 1
            self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")

        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def gen_message(self, m: tuple[str, list[tuple[str,str]]]):
        self.gen_struct(m)

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}.")
        self.write_line(f"// Do not edit directly.")
        self.write_line()
        self.write_line("using System;")
        self.write_line("using System.IO;")
        self.write_line("using System.Text;")
        self.write_line("using System.Collections.Generic;")
        self.write_line()

        if self.protocol.namespace:
            self.write_line(f"namespace {self.protocol.namespace}")
            self.write_line("{")
            self.indent_level += 1

        msg_types = [mt for mt in self.protocol.messages.keys()]

        self.write_line("public enum MessageType")
        self.write_line("{")
        self.indent_level += 1
        [self.write_line(f"{k}Type = {i+1},") for i, k in enumerate(msg_types)]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("public abstract class Message {")
        self.indent_level += 1
        self.write_line("abstract public MessageType GetMessageType();")
        self.write_line("abstract public void WriteBytes(BinaryWriter bw, bool tag);")
        self.write_line("abstract public int GetSizeInBytes();")
        self.write_line()
        self.write_line("public static Message[] ProcessRawBytes(BinaryReader br)")
        self.write_line("{")
        self.indent_level += 1
        self.write_line("List<Message> msgList = new List<Message>();")
        self.write_line("while (br.BaseStream.Position < br.BaseStream.Length)")
        self.write_line("{")
        self.indent_level += 1
        self.write_line("byte msgType = br.ReadByte();")
        self.write_line("switch (msgType)")
        self.write_line("{")
        self.indent_level += 1
        for msg_type in msg_types:
            self.write_line(f"case (byte)MessageType.{msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgList.Add({msg_type}.FromBytes(br));")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgList.Add(null);")
        self.write_line("break;")
        self.indent_level -= 1
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("if (msgList[msgList.Count-1] == null) {")
        self.indent_level += 1
        self.write_line("break;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return msgList.ToArray();")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for s in self.protocol.structs.items():
            self.gen_struct(s)

        for m in self.protocol.messages.items():
            self.gen_message(m)

        if self.protocol.namespace:
            self.indent_level -= 1
            self.write_line("}")

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



from ..protocol import Protocol, BASE_TYPE_SIZES, COLLECTION_TYPES
from ..writer import Writer
from .. import LIB_NAME, LIB_VERSION

LANGUAGE_NAME = "TypeScript"


class TypeScriptWriter(Writer):
    language_name = LANGUAGE_NAME
    default_extension = ".ts"

    def __init__(self, p: Protocol):
        super().__init__(protocol=p, tab="  ")

        for var_type in BASE_TYPE_SIZES:
            if var_type == "bool":
                self.type_mapping[var_type] = "boolean"
            elif var_type in ["int64", "uint64"]:
                self.type_mapping[var_type] = "bigint"
            else:
                # <sigh>, JavaScript
                self.type_mapping[var_type] = "number"


    def deserializer(self, var_type: str, var_name: str, parent: str = "this") -> list[str]:
        if parent:
            pref = parent + "."
        else:
            pref = ""
        label = var_name
        if label.endswith("_i]"):
            label = "i_"
        if var_type in BASE_TYPE_SIZES:
            func = None
            pre_wrap = ""
            post_wrap = ""
            off = 0
            if var_type == "uint16":
                func = "getUint16"
                off = 2
            elif var_type == "int16":
                func = "getInt16"
                off = 2
            elif var_type == "uint32":
                func = "getUint32"
                off = 4
            elif var_type == "int32":
                func = "getInt32"
                off = 4
            elif var_type == "uint64":
                func = "getBigUint64"
                off = 8
            elif var_type == "int64":
                func = "getBigInt64"
                off = 8
            elif var_type == "float":
                func = "getFloat32"
                pre_wrap = "Math.fround("
                post_wrap = ")"
                off = 4
            elif var_type == "double":
                func = "getFloat64"
                off = 8

            if func != None:
                return [
                    f"{pref}{var_name} = {pre_wrap}dv.{func}(offset, true){post_wrap};",
                    f"offset += {off};"
                ]

            if var_type == "byte":
                return [
                    f"{pref}{var_name} = dv.getUint8(offset);",
                    "offset += 1;"
                ]
            elif var_type == "bool":
                return [
                    f"const {label}Byte = dv.getUint8(offset);",
                    f"{pref}{var_name} = ({label}Byte > 0);",
                    "offset += 1;"
                ]
            if func == None:
                raise NotImplementedError(f"Type {var_type} not deserializable yet.")
        elif var_type == "string":
            return [
                f"const {label}Length = dv.getUint32(offset, true);",
                "offset += 4;",
                f"const {label}Buffer = new Uint8Array(dv.buffer, offset, {label}Length);",
                f"offset += {label}Length;",
                f"{pref}{var_name} = textDec.decode({label}Buffer);",
            ]
        elif var_type in self.protocol.structs:
            return [
                f"const {label}RetVal = {var_type}.FromBytes(dv, offset);",
                f"{pref}{var_name} = {label}RetVal.val;",
                f"offset = {label}RetVal.offset;",
            ]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"const {var_name}Length = dv.getUint32(offset, true);",
                "offset += 4;",
                f"{pref}{var_name} = Array<{self.get_var(interior)}>({var_name}Length);",
                f"for (let {var_name}_i = 0; {var_name}_i < {var_name}Length; {var_name}_i++)",
                "{"
            ]
            out += [
                self.tab + deser for deser in self.deserializer(
                    interior, f"{var_name}[{var_name}_i]", parent
                )
            ]
            out += "}"
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not deserializable yet.")


    def serializer(self, var_type: str, var_name: str, parent: str = "this") -> list[str]:
        if parent:
            pref = parent + "."
        else:
            pref = ""
        if var_type in BASE_TYPE_SIZES:
            func = None
            off = 0
            if var_type == "uint16":
                func = "setUint16"
                off = 2
            if var_type == "int16":
                func = "setInt16"
                off = 2
            if var_type == "uint32":
                func = "setUint32"
                off = 4
            if var_type == "int32":
                func = "setInt32"
                off = 4
            if var_type == "uint64":
                func = "setBigUint64"
                off = 8
            if var_type == "int64":
                func = "setBigInt64"
                off = 8
            if var_type == "float":
                func = "setFloat32"
                off = 4
            if var_type == "double":
                func = "setFloat64"
                off = 8

            if func != None:
                return [
                    f"dv.{func}(offset, {pref}{var_name}, true);",
                    f"offset += {off};"
                ]
            if var_type == "byte":
                return [
                    f"dv.setUint8(offset, {pref}{var_name});",
                    "offset += 1;"
                ]
            if var_type == "bool":
                return [
                    f"dv.setUint8(offset, {pref}{var_name} ? 1 : 0);",
                    "offset += 1;"
                ]
        elif var_type == "string":
            return [
                f"const {var_name}Buffer = textEnc.encode({pref}{var_name});",
                f"const {var_name}Arr = new Uint8Array(dv.buffer);",
                f"dv.setUint32(offset, {var_name}Buffer.byteLength, true);",
                "offset += 4;",
                f"{var_name}Arr.set({var_name}Buffer, offset);",
                f"offset += {var_name}Buffer.byteLength;",
            ]
        elif var_type in self.protocol.structs:
            return [
                f"offset = {pref}{var_name}.WriteBytes(dv, offset);"
            ]
        elif var_type[0] == "[" and var_type[-1] == "]":
            interior = var_type[1:-1]
            out = [
                f"dv.setUint32(offset, {pref}{var_name}.length, true);",
                "offset += 4;",
                f"for (let {var_name}_i = 0; {var_name}_i < {pref}{var_name}.length; {var_name}_i++) {{",
                self.tab + f"let el = {pref}{var_name}[{var_name}_i];",
            ]
            out += [self.tab + ser for ser in self.serializer(interior, "el", None)]
            out += "}"
            return out
        else:
            raise NotImplementedError(f"Type {var_type} not serializable yet.")

    def gen_measurement(self, s: tuple[str, list[tuple[str,str]]], accessor_prefix: str = "") -> tuple[list[str], int]:
        lines: list[str] = []

        accum = 0
        if self.protocol.is_simple(s[0]):
            lines.append(f"return {self.protocol.calculate_size(s[0])};")
        else:
            size_init = "let size: number = 0;"
            lines.append(size_init)

            for var_name, var_type in s[1]:
                if self.protocol.is_simple(var_type):
                    accum += self.protocol.calculate_size(var_type)
                else:
                    if var_type == "string":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"size += textEnc.encode({accessor_prefix}{var_name}).byteLength;")
                    elif var_type == "[string]":
                        accum += BASE_TYPE_SIZES["uint32"]
                        lines.append(f"for (let {var_name}_i=0; {var_name}_i < {accessor_prefix}{var_name}.length; {var_name}_i++) {{")
                        lines.append(f"{self.tab}size += {BASE_TYPE_SIZES['uint32']} + textEnc.encode({accessor_prefix}{var_name}[{var_name}_i]).byteLength;")
                        lines.append("}")
                    elif var_type[0] == "[" and var_type[-1] == "]":
                        listed_var_type = var_type[1:-1]
                        if self.protocol.is_simple(listed_var_type):
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"size += {accessor_prefix}{var_name}.length * {self.protocol.calculate_size(listed_var_type)};")
                        else:
                            accum += BASE_TYPE_SIZES["uint32"]
                            lines.append(f"for (let {var_name}_i=0; {var_name}_i < {accessor_prefix}{var_name}.length; {var_name}_i++) {{")
                            clines, caccum = self.gen_measurement((var_type, self.protocol.structs[listed_var_type]), f"{accessor_prefix}{var_name}[{var_name}_i].")
                            if clines[0] == size_init:
                                clines = clines[1:]
                            clines.append(f"size += {caccum};")
                            lines += [f"{self.tab}{l}" for l in clines]
                            lines.append("}")
                    else:
                        clines, caccum = self.gen_measurement((var_type, self.protocol.structs[var_type]), f"{accessor_prefix}{var_name}.")
                        if clines[0] == size_init:
                            clines = clines[1:]
                        lines += clines
                        accum += caccum
        return lines, accum

    def gen_struct(self, s: tuple[str, list[tuple[str,str]]]):
        is_message = False
        if s[0] in self.protocol.messages:
            is_message = True

        if is_message:
            self.write_line(f"export class {s[0]} implements Message {{")
        else:
            self.write_line(f"export class {s[0]} {{")
        self.indent_level += 1

        for var_name, var_type in s[1]:
            if var_type[0] == "[" and var_type[-1] == "]":
                self.write_line(f"{var_name}: {self.get_var(var_type[1:-1])}[] = [];")
            else:
                var_type = self.get_var(var_type)
                default_value = "0"
                if var_type == "bigint":
                    default_value = "0n"
                elif var_type == "boolean":
                    default_value = "false"
                elif var_type == "string":
                    default_value = "\"\""
                elif var_type in self.protocol.structs:
                    default_value = None
                self.write_line(f"{var_name}: {var_type}{f' = {default_value}' if default_value else ''};")
        self.write_line()

        if is_message:
            self.write_line(f"GetMessageType() : MessageType {{ return MessageType.{s[0]}Type; }}")
            self.write_line()
            self.write_line("GetSizeInBytes(): number {")
            self.indent_level += 1
            measure_lines, accumulator = self.gen_measurement(s, "this.")
            [self.write_line(s) for s in measure_lines]
            if accumulator > 0:
                self.write_line(f"size += {accumulator};")
            if len(measure_lines) > 1:
                self.write_line(f"return size;")
            self.indent_level -= 1
            self.write_line("}")
            self.write_line()
            self.write_line("WriteBytes(dv: DataView, tag: boolean, offset: number) : number {")
            self.indent_level += 1
            self.write_line("if (tag) {")
            self.indent_level += 1
            self.write_line(f"dv.setUint8(offset, MessageType.{s[0]}Type);")
            self.write_line("offset += 1;")
            self.indent_level -= 1
            self.write_line("}")
        else:
            self.write_line("WriteBytes(dv: DataView, offset: number) : number {")
            self.indent_level += 1

        for var_name, var_type in s[1]:
            [self.write_line(s) for s in self.serializer(var_type, var_name)]
        self.write_line("return offset;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line(f"static FromBytes(dv: DataView, offset: number): {{val: {s[0]}, offset: number}} {{")
        self.indent_level += 1
        if is_message:
            self.write_line("try {")
            self.indent_level += 1
        self.write_line(f"const n{s[0]} = new {self.get_var(s[0])}();")
        for var_name, var_type in s[1]:
            [self.write_line(s) for s in self.deserializer(var_type, var_name, f"n{s[0]}")]
        self.write_line(f"return {{val: n{s[0]}, offset: offset}};")
        if is_message:
            self.indent_level -= 1
            self.write_line("}")
            self.write_line("catch (RangeError) {")
            self.indent_level += 1
            self.write_line("return {val: null, offset: offset};")
            self.indent_level -= 1
            self.write_line("}")

        self.indent_level -= 1
        self.write_line("}")

        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

    def gen_message(self, m: tuple[str, list[tuple[str,str]]]):
        self.write_line("@staticImplements<MessageStatic>()")
        self.gen_struct(m)

    def generate(self) -> str:
        self.output = []

        self.write_line(f"// This file was automatically generated by {LIB_NAME} v{LIB_VERSION}.")
        self.write_line(f"// Do not edit directly.")
        self.write_line()

        self.write_line("const textDec = new TextDecoder('utf-8');")
        self.write_line("const textEnc = new TextEncoder();")
        self.write_line()

        self.write_line("export interface Message {")
        self.indent_level += 1
        self.write_line("GetMessageType() : MessageType;")
        self.write_line("WriteBytes(dv: DataView, tag: boolean, offset: number) : number;")
        self.write_line("GetSizeInBytes() : number;")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("export interface MessageStatic {")
        self.indent_level += 1
        self.write_line("new(): Message;")
        self.write_line("FromBytes(dv: DataView, offset: number): {val: Message | null, offset: number};")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("function staticImplements<T>() {")
        self.indent_level += 1
        self.write_line("return (constructor: T) => {}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        msg_types = [mt for mt in self.protocol.messages.keys()]

        self.write_line("export enum MessageType {")
        self.indent_level += 1
        [self.write_line(f"{k}Type = {i+1},") for i, k in enumerate(msg_types)]
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        self.write_line("export function ProcessRawBytes(dv: DataView, offset: number): {vals: Message[], offset: number} {")
        self.indent_level += 1
        self.write_line("const msgList: Message[] = [];")
        self.write_line("while (offset < dv.byteLength) {")
        self.indent_level += 1
        self.write_line("const msgType: number = dv.getUint8(offset);")
        self.write_line("offset += 1;")
        self.write_line("let msgRet: any = null;")
        self.write_line("switch (msgType) {")
        self.indent_level += 1
        for msg_type in msg_types:
            self.write_line(f"case MessageType.{msg_type}Type:")
            self.indent_level += 1
            self.write_line(f"msgRet = {msg_type}.FromBytes(dv, offset);")
            self.write_line("break;")
            self.indent_level -= 1
        self.write_line("default:")
        self.indent_level += 1
        self.write_line("msgRet = {val: null, offset: offset};")
        self.write_line("break;")
        self.indent_level -= 1
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("offset = msgRet.offset;")
        self.write_line("msgList.push(msgRet.val);")
        self.write_line("if (msgRet.val == null) {")
        self.indent_level += 1
        self.write_line("break;")
        self.indent_level -= 1
        self.write_line("}")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line("return {vals: msgList, offset: offset};")
        self.indent_level -= 1
        self.write_line("}")
        self.write_line()

        for s in self.protocol.structs.items():
            self.gen_struct(s)

        for m in self.protocol.messages.items():
            self.gen_message(m)

        self.write_line()
        assert self.indent_level == 0

        return "\n".join(self.output)



import os
import subprocess

import beschi.writers

CODE_OUTPUT_DIR = "out/generated"
DATA_OUTPUT_DIR = "out/data"
HARNESS_SRC_DIR = "test/_harnesses"
HARNESS_BIN_DIR = "out/executables"

def generate_for(protocol: str, output_name: str):
    for label, writer_class in beschi.writers.all_writers.items():
        out_file_dir =  os.path.join(CODE_OUTPUT_DIR, label)
        out_file_path = os.path.join(out_file_dir, f"{output_name}{writer_class.default_extension}")

        if os.path.exists(out_file_path):
            os.unlink(out_file_path)
        assert(not os.path.exists(out_file_path))

        if not os.path.exists(out_file_dir):
            os.makedirs(out_file_dir)

        subprocess.check_call(["beschi",
            "--lang", label,
            "--protocol", protocol,
            "--output", out_file_path
        ])
        assert(os.path.exists(out_file_path))

def build_for(language: str, srcfile: str, libfile: str):
    writer_class = beschi.writers.all_writers[language]
    if not os.path.exists(HARNESS_BIN_DIR):
        os.makedirs(HARNESS_BIN_DIR)
    harness_path = os.path.join(HARNESS_SRC_DIR, language)
    invoke_build = [
        "python", f"{harness_path}/builder.py",
        "--srcfile", f"{srcfile}{writer_class.default_extension}",
        "--libfile", f"{libfile}{writer_class.default_extension}",
    ]
    subprocess.check_call(invoke_build + ["--clean"])
    subprocess.check_call(invoke_build)

def run_for(language: str, srcfile: str):
    if not os.path.exists(DATA_OUTPUT_DIR):
        os.makedirs(DATA_OUTPUT_DIR)

    out_file = os.path.join(DATA_OUTPUT_DIR, f"{srcfile}.{language}.msg")
    if os.path.exists(out_file):
        os.unlink(out_file)
    assert(not os.path.exists(out_file))
    subprocess.check_call([
        os.path.join(HARNESS_BIN_DIR, f"{srcfile}_{language}"),
        "--generate", out_file
    ])
    assert(os.path.exists(out_file))

    subprocess.check_call([
        os.path.join(HARNESS_BIN_DIR, f"{srcfile}_{language}"),
        "--read", out_file
    ])



from .protocol import Protocol, COLLECTION_TYPES

DEFAULT_INDENT = "    "


class Writer:
    language_name = "[Base]"
    default_extension = ".beschi"

    def __init__(self, protocol: Protocol, tab: str = DEFAULT_INDENT):
        self.indent_level: int = 0
        self.output: list[str] = []

        self.protocol = protocol
        self.tab: str = tab

        self.type_mapping: dict[str, str] = {}
        for coll_type in COLLECTION_TYPES:
            self.type_mapping[coll_type] = coll_type
        for struct_type in self.protocol.structs:
            self.type_mapping[struct_type] = struct_type
        for msg_type in self.protocol.messages:
            self.type_mapping[msg_type] = msg_type

    # write line with current indentation level
    def write_line(self, text: str = ""):
        if len(text) == 0:
            self.output.append("")
        else:
            self.output.append(self.indent_string(text))

    # return text with current indentation level
    def indent_string(self, text: str) -> str:
        return (self.tab * self.indent_level) + text

    def get_var(self, var_type: str) -> str:
        if var_type not in self.type_mapping:
            raise NotImplementedError(f"No type called {var_type} implemented for <{type(self)}>.")
        return self.type_mapping[var_type]



import os
from setuptools import setup, find_packages

from beschi import LIB_VERSION, LIB_NAME

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md')) as f:
    DESC = f.read()

setup(
    name=LIB_NAME,
    version=LIB_VERSION,
    description='code generator for binary message passing between languages',
    long_description=DESC,
    long_description_content_type='text/markdown',
    url='https://github.com/sjml/beschi',
    author='Shane Liesegang',
    license='MIT',
    classifier=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
    ],
    packages=find_packages(
        include=['beschi', 'beschi.writers'],
        exclude=['test'], # seems to include test no matter what. ¯\_(ツ)_/¯
    ),
    install_requires=[
        'toml',
    ],
    extras_require={
        'dev': [
            'pytest',
            'build',
            'twine'
        ]
    },
    entry_points={
        'console_scripts': [
            'beschi = beschi.cli:main'
        ]
    }
)



import sys
import argparse

from .protocol import Protocol
from .writers import all_writers


def main():
    argparser = argparse.ArgumentParser(description="Generate the code for reading/writing messages from a given protocol.")
    argparser.add_argument("--lang", "-l", type=str, required=True, help="language to generate")
    argparser.add_argument("--output", "-o", type=str, default=None, help="path to output file; if omitted, will output to stdout")
    argparser.add_argument("--protocol", "-p", type=str, required=True, help="path to the protocol TOML file")

    args = argparser.parse_args()

    try:
        protocol = Protocol(args.protocol)
    except FileNotFoundError:
        sys.stderr.write(f"ERROR: No such file: {args.protocol}\n")
        sys.exit(1)
    except KeyError as ke:
        sys.stderr.write(f"ERROR: Invalid protocol file '{args.protocol}' is missing {ke}.\n")
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(f"ERROR: Invalid protocol file '{args.protocol}'.\n")
        sys.stderr.write(f"\t{e}\n")
        sys.exit(1)

    writers = [w.lower() for w in all_writers.keys()]
    if args.lang not in writers:
        sys.stderr.write("\n".join([
            "ERROR: invalid language. Valid writers are:",
            "\t" + ", ".join(writers),
            ""
        ]))
        sys.exit(1)

    writer_class = all_writers[args.lang]

    writer = writer_class(protocol)
    output = writer.generate()

    if args.output == None:
        print(output)
    else:
        # if there's any problem with creating/writing to this file, just let
        #    the regular exceptions bubble out
        with open(args.output, "w") as output_file:
            output_file.write(output)
